"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

import os
import shutil

from .config import BUILD_CACHE_PATH, CACHE_ROOT
from .core import dirname, isolate_root_cmd, open_file, running_as_root, spawn
from .exceptions import SysExit
from .i18n import translate
from .makepkg_config import MakePkgCommand, MakepkgConfig
from .pprint import print_error, print_stderr
from .version import VersionMatcher


class SrcInfo():

    _common_lines: list[str]
    _package_lines: list[str]
    path: str
    repo_path: str
    pkgbuild_path: str
    package_name: str | None
    pkgnames: list[str]

    def load_config(self) -> None:
        self.pkgnames = []
        self._common_lines = []
        self._package_lines = []
        if not os.path.exists(self.path):
            return
        destination = self._common_lines
        with open_file(self.path) as srcinfo_file:
            for line in srcinfo_file.readlines():
                if line.startswith("pkgname ="):
                    pkgname = line.split("=")[1].strip()
                    self.pkgnames.append(pkgname)
                    if pkgname == self.package_name:
                        destination = self._package_lines
                    else:
                        destination = []
                else:
                    destination.append(line)

    def __init__(
            self,
            repo_path: str | None = None,
            package_name: str | None = None,
            pkgbuild_path: str | None = None
    ) -> None:
        if repo_path:
            self.repo_path = repo_path
            self.pkgbuild_path = os.path.join(repo_path, "PKGBUILD")
        elif pkgbuild_path:
            self.pkgbuild_path = pkgbuild_path
            self.repo_path = dirname(pkgbuild_path)
        else:
            raise NotImplementedError("Either `repo_path` or `pkgbuild_path` should be set")
        self.path = os.path.join(
            self.repo_path,
            ".SRCINFO"
        )
        self.package_name = package_name
        self.load_config()

    def get_values(self, field: str, lines: list[str] | None = None) -> list[str]:
        prefix = field + " = "
        values = []
        if lines is None:
            lines = self._common_lines + self._package_lines
        for line in lines:
            if line.strip().startswith(prefix):
                values.append(line.strip().split(prefix)[1])
        return values

    def get_value(self, field: str, fallback: str | None = None) -> str | None:
        values = self.get_values(field)
        value = values[0] if values else None
        if value is None:
            return fallback
        return value

    def get_install_script(self) -> str | None:
        values = self.get_values("install")
        if values:
            return values[0]
        return None

    def _get_depends(self, field: str, lines: list[str] | None = None) -> dict[str, VersionMatcher]:
        if lines is None:
            lines = self._common_lines + self._package_lines
        carch = MakepkgConfig.get("CARCH")
        dependencies: dict[str, VersionMatcher] = {}
        for dep_line in (
                self.get_values(field, lines=lines) +
                self.get_values(f"{field}_{carch}", lines=lines)
        ):
            version_matcher = VersionMatcher(dep_line, is_pkg_deps=True)
            pkg_name = version_matcher.pkg_name
            if pkg_name not in dependencies:
                dependencies[pkg_name] = version_matcher
            else:
                dependencies[pkg_name].add_version_matcher(version_matcher)
        return dependencies

    def _get_build_depends(self, field: str) -> dict[str, VersionMatcher]:
        return self._get_depends(field=field, lines=self._common_lines)

    def get_depends(self) -> dict[str, VersionMatcher]:
        return self._get_depends("depends")

    def get_build_depends(self) -> dict[str, VersionMatcher]:
        return self._get_build_depends("depends")

    def get_build_makedepends(self) -> dict[str, VersionMatcher]:
        return self._get_build_depends("makedepends")

    def get_build_checkdepends(self) -> dict[str, VersionMatcher]:
        return self._get_build_depends("checkdepends")

    def get_version(self) -> str:
        epoch = self.get_value("epoch")
        epoch_display = (epoch + ":") if epoch else ""
        version = self.get_value("pkgver")
        release = self.get_value("pkgrel")
        return f"{epoch_display}{version}-{release}"

    def regenerate(self) -> None:
        working_directory = self.repo_path
        if running_as_root() and not self.repo_path.startswith(CACHE_ROOT):
            working_directory = os.path.join(
                BUILD_CACHE_PATH,
                "_info_" + (self.get_value("pkgbase") or "unknown")
            )
            if not os.path.exists(working_directory):
                os.mkdir(working_directory)
            shutil.copy(self.pkgbuild_path, working_directory)
        result = spawn(
            isolate_root_cmd(
                MakePkgCommand.get() + ["--printsrcinfo"] +
                ["-p", os.path.basename(self.pkgbuild_path)],
                cwd=working_directory
            ), cwd=working_directory
        )
        if result.returncode != 0 or not result.stdout_text:
            print_error(
                translate("failed to generate .SRCINFO from {}:").format(self.pkgbuild_path)
            )
            print_stderr(result.stderr_text)
            raise SysExit(5)
        with open_file(self.path, "w") as srcinfo_file:
            srcinfo_file.write(result.stdout_text)
        self.load_config()



import gzip
import json
import socket
from time import sleep
from typing import Any
from urllib import request
from urllib.error import URLError

from .args import parse_args
from .config import PikaurConfig
from .exceptions import SysExit
from .i18n import translate
from .pprint import ColorsHighlight, color_line, print_error, print_stderr
from .prompt import ask_to_continue

DEFAULT_WEB_ENCODING = "utf-8"
NOCONFIRM_RETRY_INTERVAL = 3


def read_bytes_from_url(
        url: str,
        *,
        optional: bool = False,
        autoretry: bool = True,
) -> bytes:
    args = parse_args()
    if args.print_commands:
        print_stderr(
            color_line("=> ", ColorsHighlight.cyan) + f"GET {url}"
        )
    req = request.Request(url, headers={"User-Agent": "Mozilla/5.0"})
    try:
        with request.urlopen(req) as response:  # nosec B310
            result_bytes: bytes = response.read()
            return result_bytes
    except URLError as exc:
        print_error(f"GET {url}")
        print_error("urllib: " + str(exc.reason))
        if autoretry and ask_to_continue(translate("Do you want to retry?")):  # pragma: no cover
            if args.noconfirm:
                print_stderr(
                    translate("Sleeping for {} seconds...").format(
                        NOCONFIRM_RETRY_INTERVAL
                    )
                )
                sleep(NOCONFIRM_RETRY_INTERVAL)
            return read_bytes_from_url(url, optional=optional)
        if optional:
            return b""
        raise SysExit(102) from exc


def get_unicode_from_url(url: str, *, optional: bool = False) -> str:
    result_bytes = read_bytes_from_url(url, optional=optional)
    return result_bytes.decode(DEFAULT_WEB_ENCODING)


def get_json_from_url(url: str) -> Any:
    return json.loads(get_unicode_from_url(url))


def get_gzip_from_url(url: str, *, autoretry: bool = True) -> str:
    result_bytes = read_bytes_from_url(url)
    try:
        decompressed_bytes_response = gzip.decompress(result_bytes)
    except Exception as exc:
        print_error(f"GET {url}")
        print_error("urllib: " + str(exc))
        if autoretry and ask_to_continue(translate("Do you want to retry?")):  # pragma: no cover
            args = parse_args()
            if args.noconfirm:
                print_stderr(
                    translate("Sleeping for {} seconds...").format(
                        NOCONFIRM_RETRY_INTERVAL
                    )
                )
                sleep(NOCONFIRM_RETRY_INTERVAL)
            return get_gzip_from_url(url)
        raise SysExit(102) from exc
    return decompressed_bytes_response.decode(DEFAULT_WEB_ENCODING)


class ProxyInitSocks5Error(Exception):
    pass


def init_proxy() -> None:
    net_config = PikaurConfig().network

    socks_proxy_addr = net_config.Socks5Proxy.get_str()
    if socks_proxy_addr:  # pragma: no cover
        port = 1080
        idx = socks_proxy_addr.find(":")
        if idx >= 0:
            port = int(socks_proxy_addr[idx + 1:])
            socks_proxy_addr = socks_proxy_addr[:idx]

        try:
            import socks  # type: ignore[import]  # pylint: disable=import-outside-toplevel
        except ImportError as exc:
            raise ProxyInitSocks5Error(
                translate("pikaur requires python-pysocks to use a socks5 proxy.")
            ) from exc
        socks.set_default_proxy(socks.PROXY_TYPE_SOCKS5, socks_proxy_addr, port)
        socket.socket = socks.socksocket  # type: ignore[misc]

    http_proxy_addr = net_config.AurHttpProxy.get_str()
    https_proxy_addr = net_config.AurHttpsProxy.get_str()
    if http_proxy_addr or https_proxy_addr:  # pragma: no cover
        proxies = {}
        if http_proxy_addr:
            proxies["http"] = http_proxy_addr
        if https_proxy_addr:
            proxies["https"] = https_proxy_addr
        proxy_support = request.ProxyHandler(proxies)
        opener = request.build_opener(proxy_support)
        request.install_opener(opener)


def wrap_proxy_env(cmd: list[str]) -> list[str]:  # pragma: no cover:
    """Add env with proxy to command line args."""
    net_config = PikaurConfig().network
    http_proxy_addr = net_config.AurHttpProxy.get_str()
    https_proxy_addr = net_config.AurHttpsProxy.get_str()
    if not (http_proxy_addr or https_proxy_addr):
        return cmd
    proxy_prefix = ["env", ]
    if http_proxy_addr:
        proxy_prefix.append(f"HTTP_PROXY={http_proxy_addr}")
    if https_proxy_addr:
        proxy_prefix.append(f"HTTPS_PROXY={https_proxy_addr}")
    return proxy_prefix + cmd



"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

import sys
from argparse import Namespace
from pprint import pformat
from typing import Any, NoReturn

from .argparse import ArgumentParserWithUnknowns
from .config import PikaurConfig
from .i18n import translate, translate_many

ArgSchema = list[tuple[str | None, str, None | bool | str | int]]
PossibleArgValuesTypes = list[str] | str | bool | int | None


def debug(msg: str | None = None) -> None:
    sys.stderr.write(f'{msg or ""}\n')


def debug_format(msg: Any) -> None:
    debug(pformat(msg))


PACMAN_BOOL_OPTS: ArgSchema = [
    # sync options
    ("S", "sync", None),
    ("g", "groups", None),
    ("w", "downloadonly", None),
    ("q", "quiet", False),
    ("s", "search", None),
    ("d", "nodeps", None),
    # query options
    ("Q", "query", None),
    ("o", "owns", None),
    ("l", "list", None),  # @TODO
    (None, "upgrades", None),
    # operations
    ("D", "database", None),
    ("F", "files", None),
    ("R", "remove", None),
    ("T", "deptest", None),
    ("U", "upgrade", None),
    ("V", "version", None),
    ("h", "help", None),
    # universal options
    ("v", "verbose", None),
    (None, "debug", None),
    (None, "noconfirm", None),
    (None, "needed", False),
]


def get_pikaur_bool_opts() -> ArgSchema:
    return [
        (None, "noedit", PikaurConfig().review.NoEdit.get_bool()),
        (None, "edit", None),
        (None, "namesonly", False),
        (None, "repo", None),
        ("a", "aur", None),
        (None, "keepbuild", PikaurConfig().build.KeepBuildDir.get_bool()),
        (None, "keepbuilddeps", PikaurConfig().build.KeepBuildDeps.get_bool()),
        (None, "nodiff", PikaurConfig().review.NoDiff.get_bool()),
        (None, "rebuild", None),
        (None, "dynamic-users", PikaurConfig().build.AlwaysUseDynamicUsers.get_bool()),
        ("P", "pkgbuild", None),
        (None, "install", None),
        ("G", "getpkgbuild", None),
        (None, "deps", None),
        (None, "ignore-outofdate", PikaurConfig().sync.IgnoreOutofdateAURUpgrades.get_bool()),
        (None, "pikaur-debug", None),
        (None, "hide-build-log", None),
        (None, "print-commands", PikaurConfig().ui.PrintCommands.get_bool()),
        (None, "skip-failed-build", PikaurConfig().build.SkipFailedBuild.get_bool()),
        # undocumented options:
        (None, "print-args-and-exit", None),
    ]


PACMAN_STR_OPTS: ArgSchema = [
    (None, "color", None),
    ("b", "dbpath", None),  # @TODO: pyalpm?
    ("r", "root", None),
    (None, "arch", None),  # @TODO
    (None, "cachedir", None),  # @TODO
    (None, "config", None),
    (None, "gpgdir", None),
    (None, "hookdir", None),
    (None, "logfile", None),
    (None, "print-format", None),  # @TODO
]


def get_pikaur_str_opts() -> ArgSchema:
    return [
        (None, "build-gpgdir", PikaurConfig().build.GpgDir.get_str()),
        (None, "mflags", None),
        (None, "makepkg-config", None),
        (None, "makepkg-path", None),
        (None, "pikaur-config", None),
    ]


PACMAN_COUNT_OPTS: ArgSchema = [
    ("y", "refresh", 0),
    ("u", "sysupgrade", 0),
    ("c", "clean", 0),
    ("k", "check", 0),
    ("i", "info", 0),
]


def get_pikaur_count_opts() -> ArgSchema:
    return [
        (None, "devel", 0),
    ]


PACMAN_APPEND_OPTS: ArgSchema = [
    (None, "ignore", None),
    (None, "ignoregroup", None),  # @TODO
    (None, "overwrite", None),
    (None, "assume-installed", None),  # @TODO
]


def get_pikaur_long_opts() -> list[str]:
    return [
        long_opt.replace("-", "_")
        for _short_opt, long_opt, _default in (
            get_pikaur_bool_opts() + get_pikaur_str_opts() + get_pikaur_count_opts()
        )
    ]


def get_pacman_long_opts() -> list[str]:  # pragma: no cover
    return [
        long_opt.replace("-", "_")
        for _short_opt, long_opt, _default
        in PACMAN_BOOL_OPTS + PACMAN_STR_OPTS + PACMAN_COUNT_OPTS + PACMAN_APPEND_OPTS
    ]


class IncompatibleArgumentsError(Exception):
    pass


class MissingArgumentError(Exception):
    pass


class PikaurArgs(Namespace):
    unknown_args: list[str]
    raw: list[str]
    # typehints:
    info: bool | None
    nodeps: bool | None
    owns: bool | None
    check: bool | None
    ignore: list[str]
    makepkg_config: str | None
    mflags: str | None
    makepkg_path: str | None
    quiet: bool
    sysupgrade: int
    devel: int
    namesonly: bool
    build_gpgdir: str
    needed: bool
    config: str | None
    refresh: int
    clean: int
    # positional: List[str]
    # @TODO: pylint bug:
    positional: list[str] = []

    def __getattr__(self, name: str) -> PossibleArgValuesTypes:
        result: PossibleArgValuesTypes = getattr(
            super(),
            name,
            getattr(self, name.replace("-", "_"))
        )
        return result

    def handle_the_same_letter(self) -> None:
        # pylint: disable=attribute-defined-outside-init
        if self.pkgbuild and self.info:  # handle "-i"
            self.install = self.info
            self.info = False
        if (self.getpkgbuild or self.query) and self.nodeps:  # handle "-d"
            self.deps = self.nodeps
            self.nodeps = False
        if self.sync or self.pkgbuild:
            if self.owns:  # handle "-o"
                self.repo = self.owns
                self.owns = False
            if self.check:  # handle "-k"
                self.keepbuild = True
                self.check = None
        if self.sysupgrade:  # handle "-u"
            if self.query:
                self.sysupgrade = 0
                self.upgrades = True

    def post_process_args(self) -> None:
        # pylint: disable=attribute-defined-outside-init
        self.handle_the_same_letter()

        new_ignore: list[str] = []
        for ignored in self.ignore or []:
            new_ignore += ignored.split(",")
        self.ignore = new_ignore

        if self.debug:
            self.pikaur_debug = True

        if self.pikaur_debug or self.verbose:
            self.print_commands = True

    def validate(self) -> None:
        if self.query:
            if not self.upgrades:
                for arg_name in ("aur", "repo"):
                    if getattr(self, arg_name):
                        raise MissingArgumentError("upgrades", arg_name)

    @classmethod
    def from_namespace(
            cls,
            namespace: Namespace,
            unknown_args: list[str],
            raw_args: list[str]
    ) -> "PikaurArgs":
        result = cls()
        for key, value in namespace.__dict__.items():
            setattr(result, key, value)
        result.unknown_args = unknown_args
        result.raw = raw_args
        result.post_process_args()
        return result

    @property
    def raw_without_pikaur_specific(self) -> list[str]:
        result = self.raw[:]
        for arg in ("--pikaur-debug", ):
            if arg in result:
                result.remove(arg)
        return result


class PikaurArgumentParser(ArgumentParserWithUnknowns):

    def error(self, message: str) -> NoReturn:
        exc = sys.exc_info()[1]
        if exc:
            raise exc
        super().error(message)

    def parse_pikaur_args(self, raw_args: list[str]) -> PikaurArgs:
        parsed_args, unknown_args = self.parse_known_args(raw_args)
        for arg in unknown_args[:]:
            if arg.startswith("-"):
                continue
            unknown_args.remove(arg)
            parsed_args.positional.append(arg)
        return PikaurArgs.from_namespace(
            namespace=parsed_args,
            unknown_args=unknown_args,
            raw_args=raw_args
        )

    def add_letter_andor_opt(
            self,
            action: str | None = None,
            letter: str | None = None,
            opt: str | None = None,
            default: PossibleArgValuesTypes = None
    ) -> None:
        if action:
            if letter and opt:
                self.add_argument(
                    "-" + letter, "--" + opt, action=action, default=default
                )
            elif opt:
                self.add_argument(
                    "--" + opt, action=action, default=default
                )
            elif letter:
                self.add_argument(
                    "-" + letter, action=action, default=default
                )
        else:
            if letter and opt:
                self.add_argument(
                    "-" + letter, "--" + opt, default=default
                )
            elif opt:
                self.add_argument(
                    "--" + opt, default=default
                )
            elif letter:
                self.add_argument(
                    "-" + letter, default=default
                )


class CachedArgs():

    args: PikaurArgs | None = None


def debug_args(args: list[str], parsed_args: PikaurArgs) -> NoReturn:  # pragma: no cover
    debug("Input:")
    debug_format(args)
    debug()
    parsed_dict = vars(parsed_args)
    pikaur_long_opts = get_pikaur_long_opts()
    pacman_long_opts = get_pacman_long_opts()
    pikaur_dict = {}
    pacman_dict = {}
    misc_args = {}
    for arg, value in parsed_dict.items():
        if arg in pikaur_long_opts:
            pikaur_dict[arg] = value
        elif arg in pacman_long_opts:
            pacman_dict[arg] = value
        else:
            misc_args[arg] = value
    debug("PIKAUR parsed args:")
    debug_format(pikaur_dict)
    debug()
    debug("PACMAN parsed args:")
    debug_format(pacman_dict)
    debug()
    debug("MISC parsed args:")
    debug_format(misc_args)
    debug()
    debug("Reconstructed pacman args:")
    debug_format(reconstruct_args(parsed_args))
    debug()
    debug("Reconstructed pacman args without -S:")
    debug_format(reconstruct_args(parsed_args, ignore_args=["sync"]))
    sys.exit(0)


def parse_args(args: list[str] | None = None) -> PikaurArgs:
    if CachedArgs.args:
        return CachedArgs.args
    args = args or sys.argv[1:]
    parser = PikaurArgumentParser(prog=sys.argv[0], add_help=False)

    # add some of pacman options to argparser to have them registered by pikaur
    # (they will be bypassed to pacman with the rest unrecognized args anyway)

    for letter, opt, default in PACMAN_BOOL_OPTS + get_pikaur_bool_opts():
        parser.add_letter_andor_opt(
            action="store_true", letter=letter, opt=opt, default=default
        )

    for letter, opt, default in PACMAN_COUNT_OPTS + get_pikaur_count_opts():
        parser.add_letter_andor_opt(
            action="count", letter=letter, opt=opt, default=default
        )

    for letter, opt, default in PACMAN_APPEND_OPTS:
        parser.add_letter_andor_opt(
            action="append", letter=letter, opt=opt, default=default
        )

    for letter, opt, default in PACMAN_STR_OPTS + get_pikaur_str_opts():
        parser.add_letter_andor_opt(
            action=None, letter=letter, opt=opt, default=default
        )

    parser.add_argument("positional", nargs="*")

    parsed_args = parser.parse_pikaur_args(args)

    if parsed_args.positional and "-" in parsed_args.positional and not sys.stdin.isatty():
        parsed_args.positional.remove("-")
        parsed_args.positional += [
            word
            for line in sys.stdin.readlines()
            for word in line.split()
        ]

    if parsed_args.print_args_and_exit:  # pragma: no cover
        debug_args(args, parsed_args)

    try:
        parsed_args.validate()
    except IncompatibleArgumentsError as exc:
        debug(translate(":: error: options {} can't be used together.").format(
            ", ".join([f"'--{opt}'" for opt in exc.args])
        ))
        sys.exit(1)
    except MissingArgumentError as exc:
        debug(
            translate_many(
                ":: error: option {} can't be used without {}.",
                ":: error: options {} can't be used without {}.",
                len(exc.args[1:])
            ).format(
                ", ".join([f"'--{opt}'" for opt in exc.args[1:]]),
                f"'--{exc.args[0]}'"
            )
        )
        sys.exit(1)
    CachedArgs.args = parsed_args
    return parsed_args


def reconstruct_args(parsed_args: PikaurArgs, ignore_args: list[str] | None = None) -> list[str]:
    if not ignore_args:
        ignore_args = []
    for letter, opt, _default in (
            get_pikaur_bool_opts() + get_pikaur_str_opts() + get_pikaur_count_opts()
    ):
        if letter:
            ignore_args.append(letter)
        if opt:
            ignore_args.append(opt.replace("-", "_"))
    count_args = []
    for letter, opt, _default in PACMAN_COUNT_OPTS:
        if letter:
            count_args.append(letter)
        if opt:
            count_args.append(opt.replace("-", "_"))
    reconstructed_args = {
        f"--{key}" if len(key) > 1 else f"-{key}": value
        for key, value in vars(parsed_args).items()
        if value
        if key not in ignore_args + count_args + [
            "raw", "unknown_args", "positional",  # computed members
        ] + [
            long_arg
            for _short_arg, long_arg, default in PACMAN_STR_OPTS + PACMAN_APPEND_OPTS
        ]
    }
    result = list(set(
        list(reconstructed_args.keys()) + parsed_args.unknown_args
    ))
    for args_key, value in vars(parsed_args).items():
        for letter, _opt, _default in PACMAN_COUNT_OPTS:
            opt = _opt.replace("-", "_")
            if value and opt == args_key and opt not in ignore_args and letter not in ignore_args:
                result += ["--" + opt] * value
    return result



"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

import gettext


class PikaurTranslation():

    translation: gettext.NullTranslations | None = None

    @classmethod
    def get(cls) -> gettext.NullTranslations:
        if not cls.translation:
            cls.translation = gettext.translation("pikaur", fallback=True)
        return cls.translation


def translate(msg: str) -> str:
    return PikaurTranslation.get().gettext(msg)


def translate_many(singular: str, plural: str, count: int) -> str:
    return PikaurTranslation.get().ngettext(singular, plural, count)



from distutils.core import setup

# To use a consistent encoding
import codecs
from os import path, environ
from typing import Dict, Any

if environ.get('PYPY_BUILD'):
    import setuptools  # pylint: disable=unused-import


HERE = path.abspath(path.dirname(__file__))

with codecs.open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()

SETUP_ARGS: Dict[str, Any] = dict(
    name='pikaur',  # Required
    version='1.14.6',  # Required
    description='AUR helper with minimal dependencies',  # Required
    long_description=LONG_DESCRIPTION,  # Optional
    url='https://github.com/actionless/pikaur',  # Optional
    author='Yauheni Kirylau',  # Optional
    author_email='actionless.loveless@gmail.com',  # Optional
    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Environment :: Console',

        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Other Audience',
        'Intended Audience :: System Administrators',

        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        'Operating System :: POSIX :: Linux',

        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',

        'Topic :: Software Development :: Build Tools',
        'Topic :: System :: Archiving :: Packaging',
        'Topic :: System :: Installation/Setup',
        'Topic :: Utilities',
    ],

    # Note that this is a string of words separated by whitespace, not a list.
    keywords='arch linux aur helper',

    packages=["pikaur", ],  # Required
)

EXTRA_ARGS_SETUPTOOLS = dict(
    long_description_content_type="text/markdown",
    install_requires=["pyalpm"],
    entry_points={
        'console_scripts': [
            'pikaur = pikaur.main:main'
        ]
    }
)

if environ.get('PYPY_BUILD'):
    SETUP_ARGS.update(EXTRA_ARGS_SETUPTOOLS)

setup(**SETUP_ARGS)



"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""
from .args import get_pikaur_long_opts, parse_args, reconstruct_args
from .config import PikaurConfig
from .core import spawn
from .i18n import translate
from .pprint import print_stdout


def _format_options_help(options: list[tuple[str, str, str]]) -> str:
    return "\n".join([
        "{:>5} {:<16} {}".format(
            short_opt and (short_opt + ",") or "",
            long_opt or "",
            descr if ((len(short_opt) + len(long_opt)) < 16) else f"\n{23 * ' '}{descr}"
        )
        for short_opt, long_opt, descr in options
    ])


def cli_print_help() -> None:
    args = parse_args()

    proc = spawn(
        [PikaurConfig().misc.PacmanPath.get_str(), ] +
        reconstruct_args(args, ignore_args=get_pikaur_long_opts()),
    )
    if not proc.stdout_text:
        raise RuntimeError("No response from Pacman")
    pacman_help = proc.stdout_text.replace(
        "pacman", "pikaur"
    ).replace(
        "options:", "\n" + translate("Common pacman options:")
    )
    if "--help" in pacman_help:
        pacman_help += (
            "\n" +
            translate("pikaur-specific operations:") + "\n    " +
            translate("pikaur {-P --pkgbuild}    [options] <file(s)>") + "\n    " +
            translate("pikaur {-G --getpkgbuild} [options] <package(s)>")
        )
    if args.pkgbuild:
        pacman_help = (
            translate("usage:  pikaur {-P --pkgbuild} [options] <file(s)>") + "\n\n" +
            translate(
                "All common pacman options as when doing `pacman -U <pkg_file>`. See `pacman -Uh`."
            )
        )
    if args.getpkgbuild:
        pacman_help = (
            translate("usage:  pikaur {-G --getpkgbuild} [options] <package(s)>")
        )

    pikaur_options_help: list[tuple[str, str, str]] = []
    if args.getpkgbuild:
        pikaur_options_help += [
            ("-d", "--deps", translate("download also AUR dependencies")),
        ]
    if args.pkgbuild:
        pikaur_options_help += [
            ("-i", "--install", translate("install built package")),
        ]
    if args.sync or args.query or args.pkgbuild:
        pikaur_options_help += [
            ("-a", "--aur", translate("query packages from AUR only")),
        ]
    if args.query:
        pikaur_options_help += [
            ("", "--repo", translate("query packages from repository only")),
        ]
    if args.sync or args.pkgbuild:
        pikaur_options_help += [
            ("-o", "--repo", translate("query packages from repository only")),
            ("", "--noedit", translate("don't prompt to edit PKGBUILDs and other build files")),
            ("", "--edit", translate("prompt to edit PKGBUILDs and other build files")),
            ("-k", "--keepbuild", translate("don't remove build dir after the build")),
            ("", "--rebuild", translate("always rebuild AUR packages")),
            ("", "--mflags=<--flag1>,<--flag2>", translate("cli args to pass to makepkg")),
            ("", "--makepkg-config=<path>", translate("path to custom makepkg config")),
            ("", "--makepkg-path=<path>", translate("override path to makepkg executable")),
            ("", "--pikaur-config=<path>", translate("path to custom pikaur config")),
            ("", "--dynamic-users", translate("always isolate with systemd dynamic users")),
            ("", "--build-gpgdir=<path>",
                translate("set GnuPG home directory used when validating package sources")),
            ("", "--skip-failed-build", translate("skip failed builds")),
            ("", "--hide-build-log", translate("hide build log")),
        ]
    if args.sync:
        pikaur_options_help += [
            ("", "--namesonly", translate("search only in package names")),
            ("", "--devel", translate("always sysupgrade '-git', '-svn' and other dev packages")),
            ("", "--nodiff", translate("don't prompt to show the build files diff")),
            ("", "--ignore-outofdate", translate(
                "ignore AUR packages' updates which marked 'outofdate'"
            )),
        ]

    if pikaur_options_help:  # if it's not just `pikaur --help`
        pikaur_options_help += [
            ("", "--print-commands", translate("print spawned by pikaur subshell commands")),
            ("", "--pikaur-debug", translate("show only debug messages specific to pikaur")),
        ]

    print_stdout("".join([
        pacman_help,
        "\n\n" + translate("Pikaur-specific options:") + "\n" if pikaur_options_help else "",
        _format_options_help(pikaur_options_help),
    ]))



"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

import datetime
import os
from html.parser import HTMLParser
from typing import TYPE_CHECKING, TextIO

try:
    from defusedxml.ElementTree import fromstring  # type: ignore[import]
except ModuleNotFoundError:
    from xml.etree.ElementTree import fromstring  # nosec B405

from .config import CACHE_ROOT, PikaurConfig
from .core import open_file
from .i18n import translate
from .pacman import PackageDB
from .pprint import (
    ColorsHighlight,
    bold_line,
    color_line,
    create_debug_logger,
    format_paragraph,
    print_error,
    print_stdout,
)
from .urllib import get_unicode_from_url

if TYPE_CHECKING:
    from xml.etree.ElementTree import Element  # nosec B405


DT_FORMAT = "%a, %d %b %Y %H:%M:%S %z"

_debug = create_debug_logger("news")


class News:
    URL = PikaurConfig().network.NewsUrl.get_str()
    CACHE_FILE = os.path.join(CACHE_ROOT, "last_seen_news.dat")
    _news_feed: "Element | None"

    def __init__(self) -> None:
        self._news_feed = None
        _debug("init")

    def print_news(self) -> None:
        _debug("print")
        if self._news_feed is None:
            print_error(translate("Could not fetch archlinux.org news"))
            return
        news_entry: "Element"
        first_news = True
        news_entry_to_update_last_seen_date = None
        try:  # pylint: disable=too-many-nested-blocks
            for news_entry in self._news_feed.iter("item"):
                child: "Element"
                for child in news_entry:
                    if "pubDate" in child.tag:
                        if self._is_new(str(child.text)):
                            if first_news:
                                print_stdout(
                                    "\n" +
                                    color_line(
                                        translate("There is news from archlinux.org!"),
                                        ColorsHighlight.red
                                    ) +
                                    "\n"
                                )
                            self._print_one_entry(news_entry)
                            # news are in inverse chronological order (newest first).
                            # if there is something to print, we save this date
                            # in our cache
                            if first_news:
                                first_news = False
                                news_entry_to_update_last_seen_date = news_entry
                        else:
                            # no more news
                            return
        finally:
            if news_entry_to_update_last_seen_date:
                self._update_last_seen_news(news_entry_to_update_last_seen_date)

    def fetch_latest(self) -> None:
        _debug("fetch_latest")
        str_response = get_unicode_from_url(self.URL, optional=True)
        if not str_response:
            print_error(translate("Could not fetch archlinux.org news"))
            return
        self._news_feed = fromstring(str_response)  # nosec B314

    def _get_last_seen_news_date(self) -> datetime.datetime:
        last_seen_fd: TextIO
        try:
            _debug(f"loading date from {self.CACHE_FILE}")
            with open_file(self.CACHE_FILE) as last_seen_fd:
                file_data = last_seen_fd.readline().strip()
                parsed_date = datetime.datetime.strptime(
                    file_data, DT_FORMAT
                )
                _debug(f"{file_data=}, {parsed_date=}")
                return parsed_date
        except (IOError, ValueError):
            # if file doesn't exist or corrupted,
            # this feature was run the first time
            # then we get take the date from the last installed package:
            last_pkg_date: datetime.datetime = datetime.datetime.fromtimestamp(
                PackageDB.get_last_installed_package_date()
            )
            last_pkg_date = last_pkg_date.replace(
                tzinfo=datetime.datetime.now().astimezone().tzinfo
            )
            time_formatted: str = last_pkg_date.strftime(DT_FORMAT)
            try:
                with open_file(self.CACHE_FILE, "w") as last_seen_fd:
                    last_seen_fd.write(time_formatted)
            except IOError:
                print_error(translate("Could not initialize {}").format(self.CACHE_FILE))
            return last_pkg_date

    def _is_new(self, last_online_news: str) -> bool:
        if not last_online_news:
            print_error(translate("The news feed could not be received or parsed."))
            return False
        last_online_news_date: datetime.datetime = datetime.datetime.strptime(
            last_online_news, DT_FORMAT
        )
        last_seen_news_date = self._get_last_seen_news_date()
        _debug(f"is_new, {last_online_news_date=}, {last_seen_news_date=}")
        return last_online_news_date > last_seen_news_date

    @staticmethod
    def _print_one_entry(news_entry: "Element") -> None:
        child: "Element"
        for child in news_entry:
            if "title" in child.tag:
                title = str(child.text)
            if "pubDate" in child.tag:
                pub_date = str(child.text)
            if "description" in child.tag:
                description = str(child.text)
        print_stdout(
            color_line(title, ColorsHighlight.cyan) +
            " (" + bold_line(pub_date) + ")"
        )
        print_stdout(
            format_paragraph(strip_tags(description))
        )
        print_stdout()

    def _update_last_seen_news(self, news_entry: "Element") -> None:
        child: "Element"
        for child in news_entry:
            if "pubDate" in child.tag:
                pub_date = str(child.text)
                break
        try:
            with open_file(self.CACHE_FILE, "w") as last_seen_fd:
                last_seen_fd.write(pub_date)
        except IOError:
            print_error(translate("Could not update {}").format(self.CACHE_FILE))


class MLStripper(HTMLParser):
    """HTMLParser that only removes HTML statements."""

    def error(self, message: object) -> None:
        pass

    def __init__(self) -> None:
        super().__init__()
        self.reset()
        self.strict = False
        self.convert_charrefs = True
        self.fed: list[str] = []

    def handle_data(self, data: str) -> None:
        self.fed.append(data)

    def get_data(self) -> str:
        return "".join(self.fed)


def strip_tags(html: str) -> str:
    """Removes HTML tags from a string, returns the plain string."""
    mlstripper = MLStripper()
    mlstripper.feed(html)
    return mlstripper.get_data()
